/*
 *    DrawCanvas.java
 *    Copyright (C) 2000 Gilad Mishne
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package weka.applet;

import java.awt.*;
import java.io.*;
import java.awt.event.*;
import java.applet.*;
import java.util.Vector;
import java.util.Stack;
import java.util.Random;
import weka.core.*;
import weka.optimization.*;
import weka.classifiers.*;


class DrawCanvas extends Canvas implements MouseListener 
{
    // Define the constant we will need
    public static final int SINGLE = 0;
    public static final int MULTIPLE = 1;   
    public static final int BLUE = -1;
    public static final int RED = 1;
    public DrawThread drawThr = new DrawThread();
    boolean ARFF_PARSED = false;
    boolean DRAW_UNKNOWN_REGION = false;
    int ur_x=0;
    int ur_y=0;
    int ur_sz=0;
    LocBoostApplet myParent;
    Random rand = new Random();
    // Flag to indicate the mode of operation (state)
    public int mode;
    boolean DrawAllFlag;
    // Dynamic arrays of points
    Vector redPoints, redSplit, redTest;
    Vector bluePoints, blueSplit, blueTest;
    Stack actionStk, colorStk;

    // Current color
    Color curColor;
    
    Point point;
    ArffFrame aFrame, rFrame;
    
    String arffString;
    String arffTString;
	Instances trainInstances = null;
	Instances testInstances = null;
	public boolean displayBoundaries = false;
	boolean lastColorWasBlue = false;
	boolean colorChanged = false;
	Instance inst = new Instance(3);
	
    TextField MessageText;
    Label numOfBluePoints;
    Label numOfRedPoints;
    
    public DrawCanvas(TextField mtext, Label blueCounter, Label redCounter, LocBoostApplet parent) 
    {
        myParent = parent;
        mode = MULTIPLE;
        DrawAllFlag = false;
        redPoints = new Vector();
        bluePoints = new Vector();
        redSplit = new Vector();
        blueSplit = new Vector();
        redTest = new Vector();
        blueTest = new Vector();
        actionStk = new Stack();
        colorStk = new Stack();
        MessageText = mtext;
        numOfBluePoints = blueCounter;
        numOfRedPoints = redCounter;
        setBackground(Color.white);
        addMouseListener(this);        
    }
       
    public void resetPoints()
    {
        redPoints.removeAllElements();
        bluePoints.removeAllElements();        
        redSplit.removeAllElements();
        blueSplit.removeAllElements();        
        redTest.removeAllElements();
        blueTest.removeAllElements();        
        point = null;
        DrawAllFlag = true;
        displayBoundaries = false;
        ARFF_PARSED = false;
        DRAW_UNKNOWN_REGION = false;
        // Update the number of point text field
        numOfBluePoints.setText("0");
        numOfRedPoints.setText("0");
        myParent.undoB.setVisible(false);        
        repaint();
    }
    
    void addPoints(int color, Point center)
	{
	    int numOfPoints = Integer.parseInt(myParent.numTF.getText());
	    actionStk.push(new Integer(numOfPoints));
	    colorStk.push(new Integer(color));
	    if (myParent.distribChoice.getSelectedIndex()==0)
	        normalRandomPoints(numOfPoints, Double.valueOf(myParent.sigmaTF.getText()).doubleValue(),
	                           color, center);
	    else if (myParent.distribChoice.getSelectedIndex()==1)
	        randomPoints(numOfPoints, Double.valueOf(myParent.coverTF.getText()).doubleValue(), color, center);
	}
	
	public void displayDecisionBoundaries()
	{
	    if (drawThr!=null)
	        drawThr.stop();
        drawThr = new DrawThread();
        drawThr.setPriority (Thread.MIN_PRIORITY);
        drawThr.setCanvas(this);
        drawThr.start();
	}
	
	private void drawDecisionRect(double x, double y, Graphics g, double size) 
	{
	    inst.setValue(0, x); 
        inst.setValue(1, y);                 // create dummy instance
        double arr[] = new double[2];
        double colorparam;
        try {
            arr = ((DistributionClassifier)myParent.myClassifier).distributionForInstance(inst);
            colorparam = arr[1];
            if (colorparam>1.0) colorparam=(float)1.0;
            if (colorparam<0.0) colorparam=(float)0.0;
            if (colorparam>0.5) {
                colorparam = 0.5+Math.sqrt((colorparam-0.5)/2);
            } else {
                colorparam = (colorparam*colorparam*2);
            }
            if (colorparam>1.0) colorparam=(float)1.0;
            if (colorparam<0.0) colorparam=(float)0.0;
            
           	Color tmp = new Color( 117+(int)(colorparam*121),
           	                       197-(int)(colorparam*42),
           	                       240-(int)(colorparam*80) );
            if (!(myParent.gradientCB.getState())) {
                tmp = (arr[1]<=0.5)?Color.cyan:Color.pink;
            }
            g.setColor(tmp);
            if (arr[0]<0.5) {
                colorChanged = !(lastColorWasBlue);
                lastColorWasBlue = true;
            } else {
                colorChanged = lastColorWasBlue;
                lastColorWasBlue = false;
            }
        } catch (Exception e) {e.printStackTrace(); }
        int xInt = (int)((x+0.5)*400);
        int yInt = (int)((y+0.5)*400);
        int sInt = (int)(size*400);
        g.fillRect(xInt-sInt/2,yInt-sInt/2,sInt+1,sInt+1);
    }
    
    public void randomPoints(int n, double coverage, int color, Point center)
    {
        Dimension dim = getSize();
        double newX = (double)dim.width*Math.sqrt(coverage/100);
        double newY = (double)dim.height*Math.sqrt(coverage/100);
        for(int i = 0; i < n; i++)
        {
            Point p = new Point((int)((Math.random()*newX) - 0.5*newX + center.x), 
                                (int)((Math.random()*newY) - 0.5*newY + center.y));
            switch (color) {
                case BLUE :      bluePoints.addElement(p); break;
                case RED :      redPoints.addElement(p);  break;
            }
        }
        // Set the redraw of all points
        DrawAllFlag = true;
        point = null;
        // Update the number of point text field
        switch (color) {
            case BLUE :      numOfBluePoints.setText(String.valueOf(n+Integer.parseInt(numOfBluePoints.getText()))); break;
            case RED :      numOfRedPoints.setText(String.valueOf(n+Integer.parseInt(numOfRedPoints.getText())));   break;
        }                                
        repaint();
    }
    
    // Return a random value according to the normal distribution
    public double randNorm(double mean, double sigma)
    {
        double u = rand.nextGaussian();
        return mean+u*sigma;
    }

    public void normalRandomPoints(int n, double sigma, int color, Point center)
    {
        double x,y;
        Dimension dim = getSize();
        sigma *= 40;    // a little normalizing
        int i = 0;
        while(i < n) {
            x = (randNorm(center.x, sigma));
            y = (randNorm(center.y, sigma));
            if(x > 0 && x < dim.width && y > 0 && y < dim.height) {
                Point p = new Point((int)x,(int)y);
                switch (color) {
                    case BLUE :      bluePoints.addElement(p); break;
                    case RED :      redPoints.addElement(p);  break;
                }
                i++;
            }
        }
        // Set the redraw of all points
        DrawAllFlag = true;
        point = null;
        // Update the number of point text field
        switch (color) {
            case BLUE :     numOfBluePoints.setText(String.valueOf(n+Integer.parseInt(numOfBluePoints.getText()))); break;
            case RED :      numOfRedPoints.setText(String.valueOf(n+Integer.parseInt(numOfRedPoints.getText())));   break;
        }                                
        repaint();
    }

    public void mousePressed(MouseEvent e) 
    {      
        if (myParent.undoB.isShowing()==false)
            myParent.undoB.setVisible(true);
        ARFF_PARSED = false;
        DRAW_UNKNOWN_REGION = false;
        int mask;
        switch (mode)  {
            case SINGLE:
                point = new Point(e.getX(), e.getY());
                mask = e.getModifiers();  // fix for solaris bug
		        if((mask & InputEvent.BUTTON1_MASK) != 0 ||
		        (mask & (InputEvent.BUTTON2_MASK | InputEvent.BUTTON3_MASK)) == 0) 
		        {  
			        bluePoints.addElement(point);
			        curColor = Color.blue;
			        numOfBluePoints.setText(String.valueOf(bluePoints.size()));
    		    } else {
	    		    redPoints.addElement(point);
		    	    curColor = Color.red;                 
			        numOfRedPoints.setText(String.valueOf(redPoints.size()));
		        }
                repaint();
                actionStk.push(new Integer(1));
                int c = (curColor==Color.red)?RED:BLUE;
                colorStk.push(new Integer(c));
                break;
            
            case MULTIPLE:
                Point center = new Point(e.getX(), e.getY());
                mask = e.getModifiers();  // fix for solaris bug
		        if((mask & InputEvent.BUTTON1_MASK) != 0 ||
		           (mask & (InputEvent.BUTTON2_MASK | InputEvent.BUTTON3_MASK)) == 0) {
				    addPoints(BLUE, center);
				} else {
				    addPoints(RED, center);
				}
                repaint();
                break;
        }
    }
    
    public void mouseReleased(MouseEvent e) {}
    public void mouseEntered(MouseEvent e) {}
    public void mouseExited(MouseEvent e) {}
    public void mouseClicked(MouseEvent e) {}

    void saveData() {
		try {
		    Frame f = new Frame();
		    FChooser win = new FChooser(f,"Save as...");
		    File dir;
		    try {
		        dir = new File("weka/data");
		    } catch (Exception e) { return; }

		    String[] files = dir.list();
    		sortStringArray(files);	    
		    if (files!=null)
		        for (int i=0; i<files.length; i++) 
		            win.list.addItem(files[i]);
		    win.setVisible(true);
            
		    if (win.selection==null) return;
		    File outputFile = new File("weka/data/" + win.selection);
		    File dirFile = new File("weka/data-dir");
			FileOutputStream fout = new FileOutputStream(outputFile);
			FileOutputStream dout = new FileOutputStream(dirFile);
			ObjectOutputStream oout = new ObjectOutputStream(fout);
			ObjectOutputStream doout = new ObjectOutputStream(dout);
			String[] newfiles=new String[files.length+1];
	        for (int i=0; i<files.length; i++) 
	            newfiles[i] = files[i];
	        newfiles[files.length] = win.selection;
			doout.writeObject(newfiles);
            win.dispose(); f.dispose();

			oout.writeObject(redPoints);
			oout.writeObject(bluePoints);
			oout.close();
			oout.flush();
		}
		catch( Exception e) { e.printStackTrace();	}
    }

    void loadData() {
		Frame f = new Frame();
		FChooser win = new FChooser(f,"Load...");
        String[] files = null;
        try {
		    java.net.URL dir = new java.net.URL(myParent.getCodeBase() + "weka/data-dir");
		    InputStream in = dir.openStream();
		    ObjectInputStream ois = new ObjectInputStream(in);
		    files = (String[])(ois.readObject());
	    } catch (Exception e) { e.printStackTrace(); }
		sortStringArray(files);
		if (files!=null)
		    for (int i=0; i<files.length; i++)
		        win.list.addItem(files[i]);
		win.setVisible(true);        
		if (win.selection==null) return;
	    try {
		    java.net.URL dest = new java.net.URL(myParent.getCodeBase() + "/weka/data/" + win.selection );
		    InputStream in = dest.openStream();
		    ObjectInputStream o = new ObjectInputStream(in);
    		redPoints.removeAllElements();
	        bluePoints.removeAllElements();
            redSplit.removeAllElements();
            blueSplit.removeAllElements();        
            redTest.removeAllElements();
            blueTest.removeAllElements();        	        
			redPoints = ((Vector)o.readObject());
			bluePoints = ((Vector)o.readObject());
	    } catch (Exception e) { e.printStackTrace(); }
		DrawAllFlag = true; 
		numOfBluePoints.setText(String.valueOf(bluePoints.size()));
		numOfRedPoints.setText(String.valueOf(redPoints.size()));
		repaint();
    }

    void splitData(int percent)
    {
        if ((percent>100) || (percent<0)) return;
        myParent.msgLabel.setText("Splitting data ...");
        Random rand = new Random();
        redPoints.trimToSize();
        bluePoints.trimToSize();
        int i, randPosition;
        int redSize = redPoints.size();
        int blueSize = bluePoints.size();
        int redSplitSize = (redSize*percent)/100;
        int blueSplitSize = (blueSize*percent)/100;
        
        redSplit.ensureCapacity(redSize);
        blueSplit.ensureCapacity(blueSize);
        
        //clear split vectors
        redSplit.removeAllElements();
        blueSplit.removeAllElements();
        redTest.removeAllElements();
        blueTest.removeAllElements();
        redTest.trimToSize();
        blueTest.trimToSize();
        
        for (i=0; i<redSize; i++) 
            redSplit.addElement(null);
        for (i=0; i<blueSize; i++) 
            blueSplit.addElement(null);
        i=0;
        while (i<redSplitSize) {
            randPosition = Math.abs(rand.nextInt()%redSize);
            if (redSplit.elementAt(randPosition)==null) {   // element was not picked yet
                redSplit.setElementAt((redPoints.elementAt(randPosition)), randPosition);
                i++;
            }
        }
        i=0;
        while (i<blueSplitSize) {
            randPosition = Math.abs(rand.nextInt()%blueSize);
            if (blueSplit.elementAt(randPosition)==null) {   // element was not picked yet
                blueSplit.setElementAt((bluePoints.elementAt(randPosition)), randPosition);
                i++;
            }
        }
        
        i=0;
        while (i<blueSize) {
            if (blueSplit.elementAt(i)==null)       // element is not in train set
                blueTest.addElement(bluePoints.elementAt(i));
            i++;
        }
        i=0;
        while (i<redSize) {
            if (redSplit.elementAt(i)==null)       // element is not in train set
                redTest.addElement(redPoints.elementAt(i));
            i++;
        }

        redSplit.trimToSize();
        blueSplit.trimToSize();        
        ARFF_PARSED = false;
        myParent.msgLabel.setText("");
        DrawAllFlag = true;
        repaint();
    }
 
    void parseInstances()
    {
        splitData(Integer.parseInt(myParent.dataPercentTF.getText()));
        myParent.msgLabel.setText("Parsing instances ...");
        int i;
	    StringBuffer text = new StringBuffer();
	    text.append("@relation LocBoostApplet\n" +
	                "@attribute x real\n" +
	                "@attribute y real\n" +
	                "@attribute Class {             " + BLUE + "," + RED + " }\n" +
	                "@data\n");
       	Rectangle bounds = getBounds();
       	double wid = bounds.width;
       	double hei = bounds.height;
        for (i=0 ;i < blueSplit.size(); i++) {
            Point p = (Point)blueSplit.elementAt(i);
            if (p==null) continue;
            text.append(((p.x/wid)-0.5) + "," + ((p.y/hei)-0.5) + "," + BLUE + "\n");
        }
        for (i=0 ;i < redSplit.size(); i++) {
            Point p = (Point)redSplit.elementAt(i);
            if (p==null) continue;
            text.append(((p.x/wid)-0.5) + "," + ((p.y/hei)-0.5) + "," + RED + "\n");
        }
        
	    StringBuffer textT = new StringBuffer();
	    textT.append("@relation LocBoostApplet\n" +
	                "@attribute x real\n" +
	                "@attribute y real\n" +
	                "@attribute Class { " + BLUE + "," + RED + " }\n" +
	                "@data\n");
        for (i=0 ;i < blueTest.size(); i++) {
            Point p = (Point)blueTest.elementAt(i);
            if (p==null) continue;
            textT.append(((p.x/wid)-0.5) + "," + ((p.y/hei)-0.5) + "," + BLUE + "\n");
        }
        for (i=0 ;i < redTest.size(); i++) {
            Point p = (Point)redTest.elementAt(i);
            if (p==null) continue;
            textT.append(((p.x/wid)-0.5) + "," + ((p.y/hei)-0.5) + "," + RED + "\n");
        }
        try {
            arffString = text.toString();
            arffTString = textT.toString();
            trainInstances = new Instances(arffString);
            trainInstances.setClassIndex(2);
            testInstances = new Instances(arffTString);
            testInstances.setClassIndex(2);
            if (testInstances.numInstances()==0)
                testInstances = trainInstances;
            ARFF_PARSED = true;
        } catch (Exception e) { e.printStackTrace(); }
        myParent.msgLabel.setText("");        
    }

    public void markUnknownRegion(int x, int y, int size)
    {
        DRAW_UNKNOWN_REGION = true;
        ur_x = x;
        ur_y = y;
        ur_sz = size;
    }
    
    public void undo()
    {
        if (actionStk.size()==0) return;
        int n = ((Integer)actionStk.pop()).intValue();
        int c = ((Integer)colorStk.pop()).intValue();
        int bs = bluePoints.size();
        int rs = redPoints.size();
        for (int i=1; i<=n; i++) {
            if (c==BLUE)
                bluePoints.removeElementAt(bs-i);
            else
                redPoints.removeElementAt(rs-i);
        }
        switch (c) {
            case BLUE :      numOfBluePoints.setText(String.valueOf(bluePoints.size())); break;
            case RED :      numOfRedPoints.setText(String.valueOf(redPoints.size()));   break;
        }                                
        DrawAllFlag = true;
        if (bluePoints.size()+redPoints.size()==0)
            myParent.undoB.setVisible(false);
        repaint();
    }

    public void update(Graphics g)
    {
        if(DrawAllFlag) {
            // Clear the panel surface to clearing color
            Rectangle bound = getBounds();
            g.clearRect(0,0,bound.width,bound.height);
        }
        paint(g);
    }
    
    public void paint(Graphics g) 
    {
       	Rectangle bound = getBounds();

    	g.setColor(Color.lightGray);
	    g.draw3DRect(0, 0, bound.width-1, bound.height-1, false);
	    
	    // draw grid
	    for (int x=0 ; x<bound.width; x+= bound.width/10)
	        g.drawLine(x, 0, x, bound.height);
	    for (int y=0 ; y<bound.height; y+= bound.height/10)
	        g.drawLine(0, y, bound.width, y);
	    g.drawLine(bound.width-1, 0, bound.width-1, bound.height);
	    g.drawLine(0, bound.height-1, bound.width, bound.height-1);
    	g.setColor(Color.black);
	    g.drawLine(bound.width/2, 0, bound.width/2, bound.height);
	    g.drawLine(0, bound.height/2, bound.width, bound.height/2);
	        
        if (displayBoundaries) {
            displayDecisionBoundaries();
        }
        drawInstances();
    }
    
    public void drawInstances()
    {
	    Graphics g = getGraphics();
        int l=myParent.instApprCB.getState()?4:1;
        if(DrawAllFlag)
        {
            DrawAllFlag = false;
            int nb = bluePoints.size();
            int nr = redPoints.size();
            int i;
            
            // draw all points
            // changing the context is a bit overhead but clears image
            for (int j=1 ;j<11; j++) {
                g.setColor(Color.blue);
                g.setPaintMode();
                for (i=(j-1)*nb/10 ;i < j*nb/10; i++) {
                    Point p = (Point)bluePoints.elementAt(i);
                    g.drawOval(p.x-l, p.y-l, l*2,l*2);
                }
                g.setColor(Color.red);
                g.setPaintMode();
                for (i=(j-1)*nr/10; i < j*nr/10; i++) {
                    Point p = (Point)redPoints.elementAt(i);
                    g.drawRect(p.x-l, p.y-l, l*2,l*2);
                }
            }
            
            //now draw points in split
            if ( (!redSplit.isEmpty()) && (!blueSplit.isEmpty()) ) {
                nb = blueSplit.size();
                nr = redSplit.size();
                for (int j=1 ;j<11; j++) {
                    g.setColor(Color.blue);
                    g.setPaintMode();
                    for (i=(j-1)*nb/10 ;i < j*nb/10; i++) {
                        Point p = (Point)blueSplit.elementAt(i);
                        if (p==null) continue;
                        g.fillOval(p.x-l, p.y-l, l*2,l*2);
                    }
                    g.setColor(Color.red);
                    g.setPaintMode();
                    for (i=(j-1)*nr/10; i < j*nr/10; i++) {
                        Point p = (Point)redSplit.elementAt(i);
                        if (p==null) continue;
                        g.fillRect(p.x-l, p.y-l, l*2,l*2);
                    }
                }
            }
        }
        
        if(point != null) {
            g.setColor(curColor);
            if (curColor == Color.red)
                g.drawRect(point.x-l, point.y-l, l*2,l*2);
            else
                g.drawOval(point.x-l, point.y-l+1, l*2,l*2);
        }
    }
        
public class DrawThread extends Thread {
    
  private DrawCanvas m_DrawCanvas;
  
  public void setCanvas(DrawCanvas d) {
    m_DrawCanvas = d;
  }
  
  // the thread calls this method
  public void run () {
    m_DrawCanvas.myParent.drawThr_suspended = false;
    m_DrawCanvas.myParent.pB.setLabel("Pause");
    m_DrawCanvas.myParent.pB.setVisible(true);
    m_DrawCanvas.myParent.msgLabel.setText("Marking decision boundaries ...");	    
	Graphics g = m_DrawCanvas.getGraphics();
    Dimension dim = m_DrawCanvas.getSize();
    double smoothness = Double.valueOf(m_DrawCanvas.myParent.smoothnessTF.getText()).doubleValue(); 
    if ((smoothness>100) || (smoothness <1)) return;
    if (smoothness>50) m_DrawCanvas.myParent.smoothingCB.setState(false);
    double interval = 0.5/smoothness;
    // the idea is to smooth just the places that "need" smoothing
    for (double x=-0.5+(interval/2); x<=0.5; x+=interval) {
        for (double y=-0.5+(interval/2); y<=0.5; y+=interval) {
            m_DrawCanvas.drawDecisionRect(x, y, g, interval);
            if ((m_DrawCanvas.colorChanged) && (m_DrawCanvas.myParent.smoothingCB.getState())) { 
                // rollback n-times more accurate
                double step = interval/5;
                for (double rollback_x=0; rollback_x<=interval; rollback_x+=step)
                    for (double rollback_y=0; rollback_y<=interval; rollback_y+=step) {
                        m_DrawCanvas.drawDecisionRect(x-interval+rollback_x, y-interval+rollback_y, g, step);
                }
                m_DrawCanvas.colorChanged = false;
            }
        }
        Thread.yield();
    }
    m_DrawCanvas.DrawAllFlag = true;            
    m_DrawCanvas.myParent.msgLabel.setText("");
    m_DrawCanvas.myParent.pB.setVisible(false);
    m_DrawCanvas.drawInstances();
    if (DRAW_UNKNOWN_REGION) {
        g.setColor(Color.black);
		g.drawOval(ur_x,ur_y,ur_sz,ur_sz);
        g.setColor(Color.yellow);
		g.drawOval(ur_x-1,ur_y-1,ur_sz+2,ur_sz+2);
		g.drawOval(ur_x+1,ur_y+1,ur_sz-2,ur_sz-2);
    }
    this.stop();  
  }
}  
  private void sortStringArray(String[] a) {
    for (int i=0; i<a.length-1; i++)
      for (int j=a.length-1; j>i; j--)
        if (a[j-1].compareTo(a[j])>0) {
          String tmp = a[j-1];
          a[j-1] = a[j];
          a[j] = tmp;
        }
  }

}

