/*
 *    ArrUtils.java
 *    Copyright (C) 2000 Gilad Mishne
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *  
 *
 * @author Gilad Mishne (gilad.mishne@intel.com)
 * @version $Revision: 1.0 $ 
 */
 
 package weka.optimization;

 
 
 /**
  * Performs actions on arrays/matrices. For performance reasons, no checks are done.
  */
 public class ArrUtils {
 
    public static double[] arr_constrain(double[] a, double lowerlimit, double upperlimit) {
        int n = a.length;
        double[] res = new double[n];
        for (int i=0; i<n ; i++) {
            if (a[i]<lowerlimit)
                res[i] = lowerlimit;
            else if (a[i]>upperlimit)
                res[i] = upperlimit;
            else res[i] = a[i];
        }
        return res;
    }

    public static double[] arr_constrain_normal(double[] a) {
        int n = a.length;
        double max = 0;
        boolean normalize=false;
        double[] res = new double[n];
        for (int i=0; i<n ; i++) 
            if ( (a[i]<-1) || (a[i]>1) ) normalize=true;
        if (normalize) {
            for (int i=0; i<n ; i++) 
                if (Math.abs(a[i])>max) max=Math.abs(a[i]);
            for (int i=0; i<n ; i++) 
                a[i] /= max;
        }
        return res;
    }

    public static double arr_max(double[] a) {
        int n = a.length;
        double res = Double.NEGATIVE_INFINITY;
        for (int i=0; i<n ; i++) 
            if (res<a[i])
                res = a[i];
        return res;
    }
    
    public static double[] arr(double[] a) {
        return arr_mul(a,1);
    }
    
    public static double[] arr_abs(double[] a) {
        int n = a.length;
        double[] res = new double[n];
        for (int i=0; i<n ; i++) 
            res[i] = Math.abs(a[i]);
        return res;
    }
    
    public static double[][] transpose(double[][] a) {
        int n = a.length;
        double[][] res = new double[n][n];
        for (int i=0; i<n ; i++) 
            for (int j=0; i<n ; j++) 
                res[i][j] = a[j][i];
        return res;
    }

    // multiply element by element
    public static double[][] mat_plus(double[][] m, double x) {
        int n = m.length;
        double[][] res = new double[n][n];
        for (int i=0; i<n ; i++) 
            for (int j=0; j<n ; j++) 
                res[i][j] = m[i][j] + x;
        return res;
    }

    // multiply element by element
    public static double[] arr_elem_mul(double[] a, double[] b) {
        int n = a.length;
        double[] res = new double[n];
        for (int i=0; i<n ; i++) 
            res[i] = a[i]*b[i];
        return res;
    }

    // multiply by scalar
    public static double[] arr_mul(double[] a, double x) {
        int n = a.length;
        double[] res = new double[n];
        for (int i=0; i<n ; i++) 
            res[i] = a[i]*x;
        return res;
    }

    // multiply two arrays
    public static double arr_mul(double[] a, double[] b) {
        int n = a.length;
        double res = 0;
        for (int i=0; i<n ; i++) 
            res += a[i]*b[i];
        return res;
    }

    // multiply matrix by array
    public static double[] arr_mul(double[][] mat, double[] a) {
        int n = a.length;        
        double[] res = new double[n];
        for (int i=0; i<n ; i++) 
            res[i] = arr_mul(mat[i],a);
        return res;
    }

    public static double[] arr_minus(double[] a, double[] b) {
        int n = a.length;
        double[] res = new double[n];
        for (int i=0; i<n ; i++) 
            res[i] = a[i]-b[i];
        return res;
    }

    public static double[] arr_plus(double[] a, double[] b) {
        int n = a.length;
        double[] res = new double[n];
        for (int i=0; i<n ; i++) 
            res[i] = a[i]+b[i];
        return res;
    }
            
    public static double arr_sum(double[] a) {
        int n = a.length;
        double res = 0;
        for (int i=0; i<n ; i++) 
            res += a[i];
        return res;
    }

    public static void arr_print(double[] a) {
        int n = a.length;        
        for (int i=0; i<n ; i++) 
            System.out.print(a[i] + " ");
        System.out.println();
    }
            
    public static double[] arr_sqr(double[] a) {
        int n = a.length;        
        double[] res = new double[n];
        for (int i=0; i<n ; i++) 
            res[i] = a[i]*a[i];
        return res;
    }
            
    public static double[][] eye(int n) {
        double[][] res = new double[n][n];
        for (int i=0; i<n ; i++) 
            for (int j=0; j<n; j++)
                res[i][j] = (i==j)?1:0;
        return res;
    }
            
 }